package ru.dbotthepony.kommons.collect

import java.util.function.BooleanSupplier
import java.util.stream.Stream

/**
 * Set, with values appearing and disappearing by conditions
 */
class ConditionalSet<E : Any> : Set<E> {
	private class Entry<E>(val value: E, var condition: BooleanSupplier)
	private val backing = ArrayList<Entry<E>>()
	private val mapped = HashMap<E, Entry<E>>()

	override val size: Int
		get() = backing.stream().filter { it.condition.asBoolean }.count().toInt()

	override fun contains(element: E): Boolean {
		return mapped[element]?.condition?.asBoolean ?: false
	}

	override fun containsAll(elements: Collection<E>): Boolean {
		return elements.all { contains(it) }
	}

	override fun isEmpty(): Boolean {
		return mapped.isEmpty() || mapped.values.stream().noneMatch { it.condition.asBoolean }
	}

	override fun stream(): Stream<E> {
		return backing.stream().filter { it.condition.asBoolean }.map { it.value }
	}

	override fun iterator(): Iterator<E> {
		return backing.iterator().filter { it.condition.asBoolean }.map { it.value }
	}

	fun actuallyContains(element: E): Boolean {
		return element in mapped
	}

	fun add(element: E, condition: BooleanSupplier): Boolean {
		if (element in mapped) return false
		val entry = Entry(element, condition)
		mapped[element] = entry
		backing.add(entry)
		return true
	}

	fun addFirst(element: E, condition: BooleanSupplier): Boolean {
		if (element in mapped) return false
		val entry = Entry(element, condition)
		mapped[element] = entry
		backing.add(0, entry)
		return true
	}

	fun replace(element: E, condition: BooleanSupplier) {
		val entry = mapped[element]

		if (entry != null) {
			entry.condition = condition
		} else {
			add(element, condition)
		}
	}

	fun replaceFirst(element: E, condition: BooleanSupplier) {
		val entry = mapped[element]

		if (entry != null) {
			entry.condition = condition
			backing.remove(entry)
			backing.add(0, entry)
		} else {
			addFirst(element, condition)
		}
	}

	fun remove(element: E): Boolean {
		val entry = mapped.remove(element)

		if (entry != null) {
			backing.remove(entry)
			return true
		}

		return false
	}
}
